@@ -1,27 +1,20 @@
 # -*- coding: utf-8 -*-
-"""
-kombu.async.hub
-===============
-
-Event loop implementation.
-
-"""
-from __future__ import absolute_import
+"""Event loop implementation."""
+from __future__ import absolute_import, unicode_literals
 
 import errno
 
-from collections import deque
 from contextlib import contextmanager
 from time import sleep
-from types import GeneratorType as generator
+from types import GeneratorType as generator  # noqa
 
-from amqp import promise
+from vine import Thenable, promise
 
-from kombu.five import Empty, range
+from kombu.five import Empty, python_2_unicode_compatible, range
 from kombu.log import get_logger
-from kombu.utils import cached_property, fileno
-from kombu.utils.compat import get_errno
+from kombu.utils.compat import fileno
 from kombu.utils.eventio import READ, WRITE, ERR, poll
+from kombu.utils.objects import cached_property
 
 from .timer import Timer
 
@@ -49,21 +42,25 @@ def _dummy_context(*args, **kwargs):
 
 
 def get_event_loop():
+    """Get current event loop object."""
     return _current_loop
 
 
 def set_event_loop(loop):
+    """Set the current event loop object."""
     global _current_loop
     _current_loop = loop
     return loop
 
 
+@python_2_unicode_compatible
 class Hub(object):
     """Event loop object.
 
-    :keyword timer: Specify timer object.
-
+    Arguments:
+        timer (kombu.async.Timer): Specify custom timer instance.
     """
+
     #: Flag set if reading from an fd will not block.
     READ = READ
 
@@ -84,7 +81,7 @@ class Hub(object):
         self.writers = {}
         self.on_tick = set()
         self.on_close = set()
-        self._ready = deque()
+        self._ready = set()
 
         self._running = False
         self._loop = None
@@ -143,7 +140,7 @@ class Hub(object):
                 except (MemoryError, AssertionError):
                     raise
                 except OSError as exc:
-                    if get_errno(exc) == errno.ENOMEM:
+                    if exc.errno == errno.ENOMEM:
                         raise
                     logger.error('Error in timer: %r', exc, exc_info=1)
                 except Exception as exc:
@@ -193,9 +190,10 @@ class Hub(object):
             self._loop = None
 
     def call_soon(self, callback, *args):
-        handle = promise(callback, args)
-        self._ready.append(handle)
-        return handle
+        if not isinstance(callback, Thenable):
+            callback = promise(callback, args)
+        self._ready.add(callback)
+        return callback
 
     def call_later(self, delay, callback, *args):
         return self.timer.call_after(delay, callback, args)
@@ -254,6 +252,11 @@ class Hub(object):
         self.writers.pop(fd, None)
         self.consolidate.discard(fd)
 
+    def on_callback_error(self, callback, exc):
+        logger.error(
+            'Callback %r raised exception: %r', callback, exc, exc_info=1,
+        )
+
     def create_loop(self,
                     generator=generator, sleep=sleep, min=min, next=next,
                     Empty=Empty, StopIteration=StopIteration,
@@ -274,15 +277,17 @@ class Hub(object):
                 tick_callback()
 
             while todo:
-                item = todo.popleft()
+                item = todo.pop()
                 if item:
                     item()
 
             poll_timeout = fire_timers(propagate=propagate) if scheduled else 1
+            #  print('[[[HUB]]]: %s' % (self.repr_active(),))
             if readers or writers:
                 to_consolidate = []
                 try:
                     events = poll(poll_timeout)
+                    #  print('[EVENTS]: %s' % (self.repr_events(events),))
                 except ValueError:  # Issue 882
                     raise StopIteration()
 
@@ -327,7 +332,7 @@ class Hub(object):
                         try:
                             next(cb)
                         except OSError as exc:
-                            if get_errno(exc) != errno.EBADF:
+                            if exc.errno != errno.EBADF:
                                 raise
                             hub_remove(fd)
                         except StopIteration:
@@ -353,7 +358,7 @@ class Hub(object):
 
     def repr_events(self, events):
         from .debug import repr_events
-        return repr_events(self, events)
+        return repr_events(self, events or [])
 
     @cached_property
     def scheduler(self):
