@@ -1,11 +1,5 @@
-"""
-kombu.common
-============
-
-Common Utilities.
-
-"""
-from __future__ import absolute_import
+"""Common Utilities."""
+from __future__ import absolute_import, unicode_literals
 
 import os
 import socket
@@ -20,10 +14,10 @@ from uuid import uuid4, uuid3, NAMESPACE_OID
 from amqp import RecoverableConnectionError
 
 from .entity import Exchange, Queue
-from .five import range
+from .five import bytes_if_py2, range
 from .log import get_logger
 from .serialization import registry as serializers
-from .utils import uuid
+from .utils.uuid import uuid
 
 try:
     from _thread import get_ident
@@ -54,35 +48,51 @@ def get_node_id():
 
 
 def generate_oid(node_id, process_id, thread_id, instance):
-    ent = '%x-%x-%x-%x' % (node_id, process_id, thread_id, id(instance))
+    ent = bytes_if_py2('%x-%x-%x-%x' % (
+        node_id, process_id, thread_id, id(instance)))
     return str(uuid3(NAMESPACE_OID, ent))
 
 
-def oid_from(instance):
-    return generate_oid(get_node_id(), os.getpid(), get_ident(), instance)
+def oid_from(instance, threads=True):
+    return generate_oid(
+        get_node_id(),
+        os.getpid(),
+        get_ident() if threads else 0,
+        instance,
+    )
 
 
 class Broadcast(Queue):
-    """Convenience class used to define broadcast queues.
+    """Broadcast queue.
+
+    Convenience class used to define broadcast queues.
 
     Every queue instance will have a unique name,
     and both the queue and exchange is configured with auto deletion.
 
-    :keyword name: This is used as the name of the exchange.
-    :keyword queue: By default a unique id is used for the queue
-       name for every consumer.  You can specify a custom queue
-       name here.
-    :keyword \*\*kwargs: See :class:`~kombu.Queue` for a list
-        of additional keyword arguments supported.
-
+    Arguments:
+        name (str): This is used as the name of the exchange.
+        queue (str): By default a unique id is used for the queue
+            name for every consumer.  You can specify a custom
+            queue name here.
+        **kwargs (Any): See :class:`~kombu.Queue` for a list
+            of additional keyword arguments supported.
     """
 
-    def __init__(self, name=None, queue=None, **kwargs):
+    attrs = Queue.attrs + (('queue', None),)
+
+    def __init__(self, name=None, queue=None, auto_delete=True,
+                 exchange=None, alias=None, **kwargs):
+        queue = queue or 'bcast.{0}'.format(uuid())
         return super(Broadcast, self).__init__(
-            name=queue or 'bcast.%s' % (uuid(), ),
-            **dict({'alias': name,
-                    'auto_delete': True,
-                    'exchange': Exchange(name, type='fanout')}, **kwargs))
+            alias=alias or name,
+            queue=queue,
+            name=queue,
+            auto_delete=auto_delete,
+            exchange=(exchange if exchange is not None
+                      else Exchange(name, type='fanout')),
+            **kwargs
+        )
 
 
 def declaration_cached(entity, channel):
@@ -90,7 +100,9 @@ def declaration_cached(entity, channel):
 
 
 def maybe_declare(entity, channel=None, retry=False, **retry_policy):
+    """Declare entity (cached)."""
     is_bound = entity.is_bound
+    orig = entity
 
     if not is_bound:
         assert channel
@@ -109,27 +121,30 @@ def maybe_declare(entity, channel=None, retry=False, **retry_policy):
 
     if retry:
         return _imaybe_declare(entity, declared, ident,
-                               channel, **retry_policy)
-    return _maybe_declare(entity, declared, ident, channel)
+                               channel, orig, **retry_policy)
+    return _maybe_declare(entity, declared, ident, channel, orig)
 
 
-def _maybe_declare(entity, declared, ident, channel):
-    channel = channel or entity.channel
+def _maybe_declare(entity, declared, ident, channel, orig=None):
     if not channel.connection:
         raise RecoverableConnectionError('channel disconnected')
-    entity.declare()
+    entity.declare(channel=channel)
     if declared is not None and ident:
         declared.add(ident)
+    if orig is not None:
+        orig.name = entity.name
     return True
 
 
-def _imaybe_declare(entity, declared, ident, channel, **retry_policy):
+def _imaybe_declare(entity, declared, ident, channel,
+                    orig=None, **retry_policy):
     return entity.channel.connection.client.ensure(
         entity, _maybe_declare, **retry_policy)(
-            entity, declared, ident, channel)
+            entity, declared, ident, channel, orig)
 
 
 def drain_consumer(consumer, limit=1, timeout=None, callbacks=None):
+    """Drain messages from consumer instance."""
     acc = deque()
 
     def on_message(body, message):
@@ -148,6 +163,7 @@ def drain_consumer(consumer, limit=1, timeout=None, callbacks=None):
 
 def itermessages(conn, channel, queue, limit=1, timeout=None,
                  callbacks=None, **kwargs):
+    """Iterator over messages."""
     return drain_consumer(
         conn.Consumer(queues=[queue], channel=channel, **kwargs),
         limit=limit, timeout=timeout, callbacks=callbacks,
@@ -161,18 +177,17 @@ def eventloop(conn, limit=None, timeout=None, ignore_timeouts=False):
     timeout errors (a timeout of 1 is often used in environments where
     the socket can get "stuck", and is a best practice for Kombu consumers).
 
-    **Examples**
+    ``eventloop`` is a generator.
 
-    ``eventloop`` is a generator::
+    Examples:
+        >>> from kombu.common import eventloop
 
-        from kombu.common import eventloop
-
-        def run(connection):
-            it = eventloop(connection, timeout=1, ignore_timeouts=True)
-            next(it)   # one event consumed, or timed out.
-
-            for _ in eventloop(connection, timeout=1, ignore_timeouts=True):
-                pass  # loop forever.
+        >>> def run(conn):
+        ...     it = eventloop(conn, timeout=1, ignore_timeouts=True)
+        ...     next(it)   # one event consumed, or timed out.
+        ...
+        ...     for _ in eventloop(conn, timeout=1, ignore_timeouts=True):
+        ...         pass  # loop forever.
 
     It also takes an optional limit parameter, and timeout errors
     are propagated by default::
@@ -180,11 +195,9 @@ def eventloop(conn, limit=None, timeout=None, ignore_timeouts=False):
         for _ in eventloop(connection, limit=1, timeout=1):
             pass
 
-    .. seealso::
-
+    See Also:
         :func:`itermessages`, which is an event loop bound to one or more
         consumers, that yields any messages received.
-
     """
     for i in limit and range(limit) or count():
         try:
@@ -198,16 +211,17 @@ def send_reply(exchange, req, msg,
                producer=None, retry=False, retry_policy=None, **props):
     """Send reply for request.
 
-    :param exchange: Reply exchange
-    :param req: Original request, a message with a ``reply_to`` property.
-    :param producer: Producer instance
-    :param retry: If true must retry according to ``reply_policy`` argument.
-    :param retry_policy: Retry settings.
-    :param props: Extra properties
-
+    Arguments:
+        exchange (kombu.Exchange, str): Reply exchange
+        req (~kombu.Message): Original request, a message with
+            a ``reply_to`` property.
+        producer (kombu.Producer): Producer instance
+        retry (bool): If true must retry according to
+            the ``reply_policy`` argument.
+        retry_policy (Dict): Retry settings.
+        **props (Any): Extra properties.
     """
-
-    producer.publish(
+    return producer.publish(
         msg, exchange=exchange,
         retry=retry, retry_policy=retry_policy,
         **dict({'routing_key': req.properties['reply_to'],
@@ -218,7 +232,7 @@ def send_reply(exchange, req, msg,
 
 
 def collect_replies(conn, channel, queue, *args, **kwargs):
-    """Generator collecting replies from ``queue``"""
+    """Generator collecting replies from ``queue``."""
     no_ack = kwargs.setdefault('no_ack', True)
     received = False
     try:
@@ -270,12 +284,10 @@ def ignore_errors(conn, fun=None, *args, **kwargs):
                 consumer.channel.close()
 
 
-    .. note::
-
+    Note:
         Connection and channel errors should be properly handled,
         and not ignored.  Using this function is only acceptable in a cleanup
         phase, like when a connection is lost or at shutdown.
-
     """
     if fun:
         with _ignore_errors(conn):
@@ -289,8 +301,11 @@ def revive_connection(connection, channel, on_revive=None):
 
 
 def insured(pool, fun, args, kwargs, errback=None, on_revive=None, **opts):
-    """Ensures function performing broker commands completes
-    despite intermittent connection failures."""
+    """Function wrapper to handle connection errors.
+
+    Ensures function performing broker commands completes
+    despite intermittent connection failures.
+    """
     errback = errback or _ensure_errback
 
     with pool.acquire(block=True) as conn:
@@ -308,15 +323,13 @@ def insured(pool, fun, args, kwargs, errback=None, on_revive=None, **opts):
 class QoS(object):
     """Thread safe increment/decrement of a channels prefetch_count.
 
-    :param callback: Function used to set new prefetch count,
-        e.g. ``consumer.qos`` or ``channel.basic_qos``.  Will be called
-        with a single ``prefetch_count`` keyword argument.
-    :param initial_value: Initial prefetch count value.
-
-    **Example usage**
-
-    .. code-block:: python
+    Arguments:
+        callback (Callable): Function used to set new prefetch count,
+            e.g. ``consumer.qos`` or ``channel.basic_qos``.  Will be called
+            with a single ``prefetch_count`` keyword argument.
+        initial_value (int): Initial prefetch count value..
 
+    Example:
         >>> from kombu import Consumer, Connection
         >>> connection = Connection('amqp://')
         >>> consumer = Consumer(connection)
@@ -344,10 +357,10 @@ class QoS(object):
 
 
         >>> def set_qos(prefetch_count):
-        ...     print('prefetch count now: %r' % (prefetch_count, ))
+        ...     print('prefetch count now: %r' % (prefetch_count,))
         >>> QoS(set_qos, 10)
-
     """
+
     prev = None
 
     def __init__(self, callback, initial_value):
@@ -358,9 +371,9 @@ class QoS(object):
     def increment_eventually(self, n=1):
         """Increment the value, but do not update the channels QoS.
 
-        The MainThread will be responsible for calling :meth:`update`
-        when necessary.
-
+        Note:
+            The MainThread will be responsible for calling :meth:`update`
+            when necessary.
         """
         with self._mutex:
             if self.value:
@@ -370,9 +383,9 @@ class QoS(object):
     def decrement_eventually(self, n=1):
         """Decrement the value, but do not update the channels QoS.
 
-        The MainThread will be responsible for calling :meth:`update`
-        when necessary.
-
+        Note:
+            The MainThread will be responsible for calling :meth:`update`
+            when necessary.
         """
         with self._mutex:
             if self.value:
