@@ -1,11 +1,5 @@
-"""
-kombu.transport.redis
-=====================
-
-Redis transport.
-
-"""
-from __future__ import absolute_import
+"""Redis transport."""
+from __future__ import absolute_import, unicode_literals
 
 import numbers
 import socket
@@ -15,33 +9,33 @@ from collections import namedtuple
 from contextlib import contextmanager
 from time import time
 
-from amqp import promise
-from anyjson import loads, dumps
+from vine import promise
 
 from kombu.exceptions import InconsistencyError, VersionMismatch
 from kombu.five import Empty, values, string_t
 from kombu.log import get_logger
-from kombu.utils import cached_property, uuid
+from kombu.utils.compat import register_after_fork
 from kombu.utils.eventio import poll, READ, ERR
 from kombu.utils.encoding import bytes_to_str
+from kombu.utils.json import loads, dumps
+from kombu.utils.objects import cached_property
+from kombu.utils.scheduling import cycle_by_name
 from kombu.utils.url import _parse_url
+from kombu.utils.uuid import uuid
 
 from . import virtual
 
-try:
-    from billiard.util import register_after_fork
-except ImportError:  # pragma: no cover
-    try:
-        from multiprocessing.util import register_after_fork  # noqa
-    except ImportError:
-        def register_after_fork(*args, **kwargs):  # noqa
-            pass
-
 try:
     import redis
 except ImportError:  # pragma: no cover
     redis = None     # noqa
 
+try:
+    from redis import sentinel
+except ImportError:  # pragma: no cover
+    sentinel = None  # noqa
+
+
 logger = get_logger('kombu.transport.redis')
 crit, warn = logger.critical, logger.warn
 
@@ -75,6 +69,7 @@ Probably the key ({1!r}) has been removed from the Redis database.
 
 
 def get_redis_error_classes():
+    """Return tuple of redis error classes."""
     from redis import exceptions
     # This exception suddenly changed name between redis-py versions
     if hasattr(exceptions, 'InvalidData'):
@@ -82,14 +77,14 @@ def get_redis_error_classes():
     else:
         DataError = exceptions.DataError
     return error_classes_t(
-        (virtual.Transport.connection_errors + tuple(filter(None, (
+        (virtual.Transport.connection_errors + (
             InconsistencyError,
             socket.error,
             IOError,
             OSError,
             exceptions.ConnectionError,
             exceptions.AuthenticationError,
-            getattr(exceptions, 'south_south_TimeoutError', None))))),
+            exceptions.TimeoutError)),
         (virtual.Transport.channel_errors + (
             DataError,
             exceptions.InvalidResponse,
@@ -98,16 +93,18 @@ def get_redis_error_classes():
 
 
 def get_redis_ConnectionError():
+    """Return the redis ConnectionError exception class."""
     from redis import exceptions
     return exceptions.ConnectionError
 
 
 class MutexHeld(Exception):
-    pass
+    """Raised when another party holds the lock."""
 
 
 @contextmanager
 def Mutex(client, name, expire):
+    """The Redis lock implementation (probably shaky)."""
     lock_id = uuid()
     i_won = client.setnx(name, lock_id)
     try:
@@ -132,7 +129,13 @@ def Mutex(client, name, expire):
                 pass
 
 
+def _after_fork_cleanup_channel(channel):
+    channel._after_fork()
+
+
 class QoS(virtual.QoS):
+    """Redis Ack Emulation."""
+
     restore_at_shutdown = True
 
     def __init__(self, *args, **kwargs):
@@ -143,7 +146,7 @@ class QoS(virtual.QoS):
         delivery = message.delivery_info
         EX, RK = delivery['exchange'], delivery['routing_key']
         with self.pipe_or_acquire() as pipe:
-            pipe.zadd(self.unacked_index_key, delivery_tag, time()) \
+            pipe.zadd(self.unacked_index_key, time(), delivery_tag) \
                 .hset(self.unacked_key, delivery_tag,
                       dumps([message._raw, EX, RK])) \
                 .execute()
@@ -225,6 +228,8 @@ class QoS(virtual.QoS):
 
 
 class MultiChannelPoller(object):
+    """Async I/O poller for Redis transport."""
+
     eventflags = READ | ERR
 
     #: Set by :meth:`get` while reading from the socket.
@@ -280,20 +285,24 @@ class MultiChannelPoller(object):
     def _unregister(self, channel, client, type):
         self.poller.unregister(self._chan_to_sock[(channel, client, type)])
 
+    def _client_registered(self, channel, client, cmd):
+        if getattr(client, 'connection', None) is None:
+            client.connection = client.connection_pool.get_connection('_')
+        return (client.connection._sock is not None and
+                (channel, client, cmd) in self._chan_to_sock)
+
     def _register_BRPOP(self, channel):
-        """enable BRPOP mode for channel."""
+        """Enable BRPOP mode for channel."""
         ident = channel, channel.client, 'BRPOP'
-        if channel.client.connection._sock is None or \
-                ident not in self._chan_to_sock:
+        if not self._client_registered(channel, channel.client, 'BRPOP'):
             channel._in_poll = False
             self._register(*ident)
-
         if not channel._in_poll:  # send BRPOP
             channel._brpop_start()
 
     def _register_LISTEN(self, channel):
-        """enable LISTEN mode for channel."""
-        if channel.subclient.connection._sock is None:
+        """Enable LISTEN mode for channel."""
+        if not self._client_registered(channel, channel.subclient, 'LISTEN'):
             channel._in_listen = False
             self._register(channel, channel.subclient, 'LISTEN')
         if not channel._in_listen:
@@ -323,12 +332,9 @@ class MultiChannelPoller(object):
                 )
 
     def on_readable(self, fileno):
-        try:
-            chan, type = self._fd_to_chan[fileno]
-        except KeyError:
-            return
+        chan, type = self._fd_to_chan[fileno]
         if chan.qos.can_consume():
-            return chan.handlers[type]()
+            chan.handlers[type]()
 
     def handle_event(self, fileno, event):
         if event & READ:
@@ -337,7 +343,7 @@ class MultiChannelPoller(object):
             chan, type = self._fd_to_chan[fileno]
             chan._poll_error(type)
 
-    def get(self, timeout=None):
+    def get(self, callback, timeout=None):
         self._in_protected_read = True
         try:
             for channel in self._channels:
@@ -348,15 +354,14 @@ class MultiChannelPoller(object):
                     self._register_LISTEN(channel)
 
             events = self.poller.poll(timeout)
-            for fileno, event in events or []:
-                ret = self.handle_event(fileno, event)
-                if ret:
-                    return ret
-
+            if events:
+                for fileno, event in events:
+                    ret = self.handle_event(fileno, event)
+                    if ret:
+                        return
             # - no new data, so try to restore messages.
             # - reset active redis commands.
             self.maybe_restore_messages()
-
             raise Empty()
         finally:
             self._in_protected_read = False
@@ -374,6 +379,8 @@ class MultiChannelPoller(object):
 
 
 class Channel(virtual.Channel):
+    """Redis Channel."""
+
     QoS = QoS
 
     _client = None
@@ -395,22 +402,54 @@ class Channel(virtual.Channel):
     visibility_timeout = 3600   # 1 hour
     priority_steps = PRIORITY_STEPS
     socket_timeout = None
+    socket_connect_timeout = None
+    socket_keepalive = None
+    socket_keepalive_options = None
     max_connections = 10
-    #: Transport option to enable disable fanout keyprefix.
-    #: Should be enabled by default, but that is not
-    #: backwards compatible.  Can also be string, in which
-    #: case it changes the default prefix ('/{db}.') into to something
-    #: else.  The prefix must include a leading slash and a trailing dot.
-    fanout_prefix = False
+    #: Transport option to disable fanout keyprefix.
+    #: Can also be string, in which case it changes the default
+    #: prefix ('/{db}.') into to something else.  The prefix must
+    #: include a leading slash and a trailing dot.
+    #:
+    #: Enabled by default since Kombu 4.x.
+    #: Disable for backwards compatibility with Kombu 3.x.
+    fanout_prefix = True
 
     #: If enabled the fanout exchange will support patterns in routing
     #: and binding keys (like a topic exchange but using PUB/SUB).
-    #: This will be enabled by default in a future version.
-    fanout_patterns = False
+    #:
+    #: Enabled by default since Kombu 4.x.
+    #: Disable for backwards compatibility with Kombu 3.x.
+    fanout_patterns = True
+
+    #: Order in which we consume from queues.
+    #:
+    #: Can be either string alias, or a cycle strategy class
+    #:
+    #: - ``round_robin``
+    #:   (:class:`~kombu.utils.scheduling.round_robin_cycle`).
+    #:
+    #:    Make sure each queue has an equal opportunity to be consumed from.
+    #:
+    #: - ``sorted``
+    #:   (:class:`~kombu.utils.scheduling.sorted_cycle`).
+    #:
+    #:    Consume from queues in alphabetical order.
+    #:    If the first queue in the sorted list always contains messages,
+    #:    then the rest of the queues will never be consumed from.
+    #:
+    #: - ``priority``
+    #:   (:class:`~kombu.utils.scheduling.priority_cycle`).
+    #:
+    #:    Consume from queues in original order, so that if the first
+    #:    queue always contains messages, the rest of the queues
+    #:    in the list will never be consumed from.
+    #:
+    #: The default is to consume from queues in round robin.
+    queue_order_strategy = 'round_robin'
 
     _async_pool = None
     _pool = None
-    _disconnecting_pools = False
 
     from_transport_options = (
         virtual.Channel.from_transport_options +
@@ -424,10 +463,16 @@ class Channel(virtual.Channel):
          'fanout_prefix',
          'fanout_patterns',
          'socket_timeout',
+         'socket_connect_timeout',
+         'socket_keepalive',
+         'socket_keepalive_options',
+         'queue_order_strategy',
          'max_connections',
          'priority_steps')  # <-- do not add comma here!
     )
 
+    connection_class = redis.Connection if redis else None
+
     def __init__(self, *args, **kwargs):
         super_ = super(Channel, self)
         super_.__init__(*args, **kwargs)
@@ -435,9 +480,8 @@ class Channel(virtual.Channel):
         if not self.ack_emulation:  # disable visibility timeout
             self.QoS = virtual.QoS
 
-        self._queue_cycle = []
-        self.AsyncClient = self._get_async_client()
-        self.Client = redis.Redis
+        self._queue_cycle = cycle_by_name(self.queue_order_strategy)()
+        self.Client = self._get_client()
         self.ResponseError = self._get_response_error()
         self.active_fanout_queues = set()
         self.auto_delete_queues = set()
@@ -454,7 +498,7 @@ class Channel(virtual.Channel):
 
         # Evaluate connection.
         try:
-            self.client.info()
+            self.client.ping()
         except Exception:
             self._disconnect_pools()
             raise
@@ -464,31 +508,31 @@ class Channel(virtual.Channel):
         # are still waiting for data.
         self.connection_errors = self.connection.connection_errors
 
-        register_after_fork(self, self._after_fork)
+        if register_after_fork is not None:
+            register_after_fork(self, _after_fork_cleanup_channel)
 
     def _after_fork(self):
         self._disconnect_pools()
 
     def _disconnect_pools(self):
-        if not self._disconnecting_pools:
-            self._disconnecting_pools = True
-            try:
-                if self._async_pool is not None:
-                    self._async_pool.disconnect()
-                if self._pool is not None:
-                    self._pool.disconnect()
-                self._async_pool = self._pool = None
-            finally:
-                self._disconnecting_pools = False
+        pool = self._pool
+        async_pool = self._async_pool
+
+        self._async_pool = self._pool = None
+
+        if pool is not None:
+            pool.disconnect()
+
+        if async_pool is not None:
+            async_pool.disconnect()
 
     def _on_connection_disconnect(self, connection):
-        self._in_poll = False
-        self._in_listen = False
+        if self._in_poll is connection:
+            self._in_poll = None
+        if self._in_listen is connection:
+            self._in_listen = None
         if self.connection and self.connection.cycle:
             self.connection.cycle._on_connection_disconnect(connection)
-        self._disconnect_pools()
-        if not self._closing:
-            raise get_redis_ConnectionError()
 
     def _do_restore_message(self, payload, exchange, routing_key,
                             client=None, leftmost=False):
@@ -527,19 +571,30 @@ class Channel(virtual.Channel):
             self.active_fanout_queues.add(queue)
             self._fanout_to_queue[exchange] = queue
         ret = super(Channel, self).basic_consume(queue, *args, **kwargs)
-        self._update_cycle()
+
+        # Update fair cycle between queues.
+        #
+        # We cycle between queues fairly to make sure that
+        # each queue is equally likely to be consumed from,
+        # so that a very busy queue will not block others.
+        #
+        # This works by using Redis's `BRPOP` command and
+        # by rotating the most recently used queue to the
+        # and of the list.  See Kombu github issue #166 for
+        # more discussion of this method.
+        self._update_queue_cycle()
         return ret
 
     def basic_cancel(self, consumer_tag):
         # If we are busy reading messages we may experience
         # a race condition where a message is consumed after
-        # cancelling, so we must delay this operation until reading
+        # canceling, so we must delay this operation until reading
         # is complete (Issue celery/celery#1773).
         connection = self.connection
         if connection:
             if connection.cycle._in_protected_read:
                 return connection.cycle.after_read.add(
-                    promise(self._basic_cancel, (consumer_tag, )),
+                    promise(self._basic_cancel, (consumer_tag,)),
                 )
             return self._basic_cancel(consumer_tag)
 
@@ -560,7 +615,7 @@ class Channel(virtual.Channel):
         except KeyError:
             pass
         ret = super(Channel, self).basic_cancel(consumer_tag)
-        self._update_cycle()
+        self._update_queue_cycle()
         return ret
 
     def _get_publish_topic(self, exchange, routing_key):
@@ -580,40 +635,50 @@ class Channel(virtual.Channel):
         c = self.subclient
         if c.connection._sock is None:
             c.connection.connect()
-        self._in_listen = True
+        self._in_listen = c.connection
         c.psubscribe(keys)
 
     def _unsubscribe_from(self, queue):
         topic = self._get_subscribe_topic(queue)
         c = self.subclient
-        should_disconnect = False
-        if c.connection._sock is None:
-            c.connection.connect()
-            should_disconnect = True
-        try:
+        if c.connection and c.connection._sock:
             c.unsubscribe([topic])
-        finally:
-            if should_disconnect and c.connection:
-                c.connection.disconnect()
 
     def _handle_message(self, client, r):
         if bytes_to_str(r[0]) == 'unsubscribe' and r[2] == 0:
             client.subscribed = False
-        elif bytes_to_str(r[0]) == 'pmessage':
-            return {'type':    r[0], 'pattern': r[1],
-                    'channel': r[2], 'data':    r[3]}
+            return
+
+        if bytes_to_str(r[0]) == 'pmessage':
+            type, pattern, channel, data = r[0], r[1], r[2], r[3]
         else:
-            return {'type':    r[0], 'pattern': None,
-                    'channel': r[1], 'data':    r[2]}
+            type, pattern, channel, data = r[0], None, r[1], r[2]
+        return {
+            'type': type,
+            'pattern': pattern,
+            'channel': channel,
+            'data': data,
+        }
 
     def _receive(self):
         c = self.subclient
+        ret = []
+        try:
+            ret.append(self._receive_one(c))
+        except Empty:
+            pass
+        if c.connection is not None:
+            while c.connection.can_read(timeout=0):
+                ret.append(self._receive_one(c))
+        return any(ret)
+
+    def _receive_one(self, c):
         response = None
         try:
             response = c.parse_response()
         except self.connection_errors:
-            self._in_listen = False
-            raise Empty()
+            self._in_listen = None
+            raise
         if response is not None:
             payload = self._handle_message(c, response)
             if bytes_to_str(payload['type']).endswith('message'):
@@ -628,16 +693,17 @@ class Channel(virtual.Channel):
                              channel, repr(payload)[:4096], exc_info=1)
                         raise Empty()
                     exchange = channel.split('/', 1)[0]
-                    return message, self._fanout_to_queue[exchange]
-        raise Empty()
+                    self.connection._deliver(
+                        message, self._fanout_to_queue[exchange])
+                    return True
 
     def _brpop_start(self, timeout=1):
-        queues = self._consume_cycle()
+        queues = self._queue_cycle.consume(len(self.active_queues))
         if not queues:
             return
-        keys = [self._q_for_pri(queue, pri) for pri in PRIORITY_STEPS
+        keys = [self._q_for_pri(queue, pri) for pri in self.priority_steps
                 for queue in queues] + [timeout or 0]
-        self._in_poll = True
+        self._in_poll = self.client.connection
         self.client.connection.send_command('BRPOP', *keys)
 
     def _brpop_read(self, **options):
@@ -650,16 +716,17 @@ class Channel(virtual.Channel):
                 # if there's a ConnectionError, disconnect so the next
                 # iteration will reconnect automatically.
                 self.client.connection.disconnect()
-                raise Empty()
+                raise
             if dest__item:
                 dest, item = dest__item
                 dest = bytes_to_str(dest).rsplit(self.sep, 1)[0]
-                self._rotate_cycle(dest)
-                return loads(bytes_to_str(item)), dest
+                self._queue_cycle.rotate(dest)
+                self.connection._deliver(loads(bytes_to_str(item)), dest)
+                return True
             else:
                 raise Empty()
         finally:
-            self._in_poll = False
+            self._in_poll = None
 
     def _poll_error(self, type, **options):
         if type == 'LISTEN':
@@ -669,7 +736,7 @@ class Channel(virtual.Channel):
 
     def _get(self, queue):
         with self.conn_or_acquire() as client:
-            for pri in PRIORITY_STEPS:
+            for pri in self.priority_steps:
                 item = client.rpop(self._q_for_pri(queue, pri))
                 if item:
                     return loads(bytes_to_str(item))
@@ -678,7 +745,7 @@ class Channel(virtual.Channel):
     def _size(self, queue):
         with self.conn_or_acquire() as client:
             with client.pipeline() as pipe:
-                for pri in PRIORITY_STEPS:
+                for pri in self.priority_steps:
                     pipe = pipe.llen(self._q_for_pri(queue, pri))
                 sizes = pipe.execute()
                 return sum(size for size in sizes
@@ -694,11 +761,8 @@ class Channel(virtual.Channel):
 
     def _put(self, queue, message, **kwargs):
         """Deliver message."""
-        try:
-            pri = max(min(int(
-                message['properties']['delivery_info']['priority']), 9), 0)
-        except (TypeError, ValueError, KeyError):
-            pri = 0
+        pri = self._get_message_priority(message, reverse=False)
+
         with self.conn_or_acquire() as client:
             client.lpush(self._q_for_pri(queue, pri), dumps(message))
 
@@ -721,27 +785,27 @@ class Channel(virtual.Channel):
                 exchange, routing_key.replace('#', '*'),
             )
         with self.conn_or_acquire() as client:
-            client.sadd(self.keyprefix_queue % (exchange, ),
+            client.sadd(self.keyprefix_queue % (exchange,),
                         self.sep.join([routing_key or '',
                                        pattern or '',
                                        queue or '']))
 
-    def _delete(self, queue, exchange, routing_key, pattern, *args):
+    def _delete(self, queue, exchange, routing_key, pattern, *args, **kwargs):
         self.auto_delete_queues.discard(queue)
-        with self.conn_or_acquire() as client:
-            client.srem(self.keyprefix_queue % (exchange, ),
+        with self.conn_or_acquire(client=kwargs.get('client')) as client:
+            client.srem(self.keyprefix_queue % (exchange,),
                         self.sep.join([routing_key or '',
                                        pattern or '',
                                        queue or '']))
             with client.pipeline() as pipe:
-                for pri in PRIORITY_STEPS:
+                for pri in self.priority_steps:
                     pipe = pipe.delete(self._q_for_pri(queue, pri))
                 pipe.execute()
 
     def _has_queue(self, queue, **kwargs):
         with self.conn_or_acquire() as client:
             with client.pipeline() as pipe:
-                for pri in PRIORITY_STEPS:
+                for pri in self.priority_steps:
                     pipe = pipe.exists(self._q_for_pri(queue, pri))
                 return any(pipe.execute())
 
@@ -756,7 +820,7 @@ class Channel(virtual.Channel):
     def _purge(self, queue):
         with self.conn_or_acquire() as client:
             with client.pipeline() as pipe:
-                for pri in PRIORITY_STEPS:
+                for pri in self.priority_steps:
                     priq = self._q_for_pri(queue, pri)
                     pipe = pipe.llen(priq).delete(priq)
                 sizes = pipe.execute()
@@ -764,25 +828,27 @@ class Channel(virtual.Channel):
 
     def close(self):
         self._closing = True
-        self._disconnect_pools()
         if not self.closed:
             # remove from channel poller.
             self.connection.cycle.discard(self)
 
             # delete fanout bindings
-            for queue in self._fanout_queues:
-                if queue in self.auto_delete_queues:
-                    self.queue_delete(queue)
-
+            client = self.__dict__.get('client')  # only if property cached
+            if client is not None:
+                for queue in self._fanout_queues:
+                    if queue in self.auto_delete_queues:
+                        self.queue_delete(queue, client=client)
+            self._disconnect_pools()
             self._close_clients()
-
         super(Channel, self).close()
 
     def _close_clients(self):
         # Close connections
         for attr in 'client', 'subclient':
             try:
-                self.__dict__[attr].connection.disconnect()
+                client = self.__dict__[attr]
+                connection, client.connection = client.connection, None
+                connection.disconnect()
             except (KeyError, AttributeError, self.ResponseError):
                 pass
 
@@ -801,22 +867,44 @@ class Channel(virtual.Channel):
                     ))
         return vhost
 
+    def _filter_tcp_connparams(self, socket_keepalive=None,
+                               socket_keepalive_options=None, **params):
+        return params
+
     def _connparams(self, async=False):
         conninfo = self.connection.client
-        connparams = {'host': conninfo.hostname or '127.0.0.1',
-                      'port': conninfo.port or DEFAULT_PORT,
-                      'virtual_host': conninfo.virtual_host,
-                      'password': conninfo.password,
-                      'max_connections': self.max_connections,
-                      'socket_timeout': self.socket_timeout}
+        connparams = {
+            'host': conninfo.hostname or '127.0.0.1',
+            'port': conninfo.port or self.connection.default_port,
+            'virtual_host': conninfo.virtual_host,
+            'password': conninfo.password,
+            'max_connections': self.max_connections,
+            'socket_timeout': self.socket_timeout,
+            'socket_connect_timeout': self.socket_connect_timeout,
+            'socket_keepalive': self.socket_keepalive,
+            'socket_keepalive_options': self.socket_keepalive_options,
+        }
+        if conninfo.ssl:
+            # Connection(ssl={}) must be a dict containing the keys:
+            # 'ssl_cert_reqs', 'ssl_ca_certs', 'ssl_certfile', 'ssl_keyfile'
+            try:
+                connparams.update(conninfo.ssl)
+                connparams['connection_class'] = redis.SSLConnection
+            except TypeError:
+                pass
         host = connparams['host']
         if '://' in host:
-            scheme, _, _, _, password, path, query = _parse_url(host)
+            scheme, _, _, _, _, path, query = _parse_url(host)
             if scheme == 'socket':
+                connparams = self._filter_tcp_connparams(**connparams)
                 connparams.update({
                     'connection_class': redis.UnixDomainSocketConnection,
-                    'path': '/' + path,
-                    'password': password}, **query)
+                    'path': '/' + path}, **query)
+
+                connparams.pop('socket_connect_timeout', None)
+                connparams.pop('socket_keepalive', None)
+                connparams.pop('socket_keepalive_options', None)
+
             connparams.pop('host', None)
             connparams.pop('port', None)
         connparams['db'] = self._prepare_virtual_host(
@@ -825,7 +913,7 @@ class Channel(virtual.Channel):
         channel = self
         connection_cls = (
             connparams.get('connection_class') or
-            redis.Connection
+            self.connection_class
         )
 
         if async:
@@ -833,13 +921,15 @@ class Channel(virtual.Channel):
                 def disconnect(self):
                     super(Connection, self).disconnect()
                     channel._on_connection_disconnect(self)
-            connparams['connection_class'] = Connection
+            connection_cls = Connection
+
+        connparams['connection_class'] = connection_cls
 
         return connparams
 
     def _create_client(self, async=False):
         if async:
-            return self.AsyncClient(connection_pool=self.async_pool)
+            return self.Client(connection_pool=self.async_pool)
         return self.Client(connection_pool=self.pool)
 
     def _get_pool(self, async=False):
@@ -847,32 +937,19 @@ class Channel(virtual.Channel):
         self.keyprefix_fanout = self.keyprefix_fanout.format(db=params['db'])
         return redis.ConnectionPool(**params)
 
-    def _get_async_client(self):
-        if redis.VERSION < (2, 4, 4):
+    def _get_client(self):
+        if redis.VERSION < (2, 10, 0):
             raise VersionMismatch(
-                'Redis transport requires redis-py versions 2.4.4 or later. '
+                'Redis transport requires redis-py versions 2.10.0 or later. '
                 'You have {0.__version__}'.format(redis))
-
-        # AsyncRedis maintains a connection attribute on it's instance and
-        # uses that when executing commands
-        # This was added after redis-py was changed.
-        class AsyncRedis(redis.Redis):  # pragma: no cover
-
-            def __init__(self, *args, **kwargs):
-                super(AsyncRedis, self).__init__(*args, **kwargs)
-                self.connection = self.connection_pool.get_connection('_')
-
-        return AsyncRedis
+        return redis.StrictRedis
 
     @contextmanager
     def conn_or_acquire(self, client=None):
         if client:
             yield client
         else:
-            if self._in_poll:
-                yield self._create_client()
-            else:
-                yield self.client
+            yield self._create_client()
 
     @property
     def pool(self):
@@ -895,38 +972,10 @@ class Channel(virtual.Channel):
     def subclient(self):
         """Pub/Sub connection used to consume fanout queues."""
         client = self._create_client(async=True)
-        pubsub = client.pubsub()
-        pool = pubsub.connection_pool
-        pubsub.connection = pool.get_connection('pubsub', pubsub.shard_hint)
-        return pubsub
+        return client.pubsub()
 
-    def _update_cycle(self):
-        """Update fair cycle between queues.
-
-        We cycle between queues fairly to make sure that
-        each queue is equally likely to be consumed from,
-        so that a very busy queue will not block others.
-
-        This works by using Redis's `BRPOP` command and
-        by rotating the most recently used queue to the
-        and of the list.  See Kombu github issue #166 for
-        more discussion of this method.
-
-        """
-        self._queue_cycle = list(self.active_queues)
-
-    def _consume_cycle(self):
-        """Get a fresh list of queues from the queue cycle."""
-        active = len(self.active_queues)
-        return self._queue_cycle[0:active]
-
-    def _rotate_cycle(self, used):
-        """Move most recently used queue to end of list."""
-        cycle = self._queue_cycle
-        try:
-            cycle.append(cycle.pop(cycle.index(used)))
-        except ValueError:
-            pass
+    def _update_queue_cycle(self):
+        self._queue_cycle.update(self.active_queues)
 
     def _get_response_error(self):
         from redis import exceptions
@@ -935,19 +984,25 @@ class Channel(virtual.Channel):
     @property
     def active_queues(self):
         """Set of queues being consumed from (excluding fanout queues)."""
-        return set(queue for queue in self._active_queues
-                   if queue not in self.active_fanout_queues)
+        return {queue for queue in self._active_queues
+                if queue not in self.active_fanout_queues}
 
 
 class Transport(virtual.Transport):
+    """Redis Transport."""
+
     Channel = Channel
 
     polling_interval = None  # disable sleep between unsuccessful polls.
     default_port = DEFAULT_PORT
-    supports_ev = True
     driver_type = 'redis'
     driver_name = 'redis'
 
+    implements = virtual.Transport.implements.extend(
+        async=True,
+        exchange_type=frozenset(['direct', 'topic', 'fanout'])
+    )
+
     def __init__(self, *args, **kwargs):
         if redis is None:
             raise ImportError('Missing redis library (pip install redis)')
@@ -981,15 +1036,66 @@ class Transport(virtual.Transport):
 
     def on_readable(self, fileno):
         """Handle AIO event for one of our file descriptors."""
-        item = self.cycle.on_readable(fileno)
-        if item:
-            message, queue = item
-            if not queue or queue not in self._callbacks:
-                raise KeyError(
-                    'Message for queue {0!r} without consumers: {1}'.format(
-                        queue, message))
-            self._callbacks[queue](message)
+        self.cycle.on_readable(fileno)
 
     def _get_errors(self):
         """Utility to import redis-py's exceptions at runtime."""
         return get_redis_error_classes()
+
+
+class SentinelChannel(Channel):
+    """Channel with explicit Redis Sentinel knowledge.
+
+    Broker url is supposed to look like:
+
+    sentinel://0.0.0.0:26379;sentinel://0.0.0.0:26380/...
+
+    where each sentinel is separated by a `;`.  Multiple sentinels are handled
+    by :class:`kombu.Connection` constructor, and placed in the alternative
+    list of servers to connect to in case of connection failure.
+
+    Other arguments for the sentinel should come from the transport options
+    (see :method:`Celery.connection` which is in charge of creating the
+    `Connection` object).
+
+    You must provide at least one option in Transport options:
+     * `service_name` - name of the redis group to poll
+    """
+
+    from_transport_options = Channel.from_transport_options + (
+        'master_name',
+        'min_other_sentinels',
+        'sentinel_kwargs')
+
+    connection_class = sentinel.SentinelManagedConnection if sentinel else None
+
+    def _sentinel_managed_pool(self, async=False):
+        connparams = self._connparams(async)
+
+        additional_params = connparams.copy()
+
+        additional_params.pop('host', None)
+        additional_params.pop('port', None)
+
+        sentinel_inst = sentinel.Sentinel(
+            [(connparams['host'], connparams['port'])],
+            min_other_sentinels=getattr(self, 'min_other_sentinels', 0),
+            sentinel_kwargs=getattr(self, 'sentinel_kwargs', {}),
+            **additional_params)
+
+        master_name = getattr(self, 'master_name', None)
+
+        return sentinel_inst.master_for(
+            master_name,
+            self.Client,
+        ).connection_pool
+
+    def _get_pool(self, async=False):
+        return self._sentinel_managed_pool(async)
+
+
+class SentinelTransport(Transport):
+    """Redis Sentinel Transport."""
+
+    default_port = 26379
+    Channel = SentinelChannel
